import numpy as np
import pandas as pd
from math import sqrt
from pyrsm import ifelse
from itertools import compress
from scipy import stats


def varprop(x, na=True):
    """
    Calculate the variance for a proportion

    Parameters
    ----------
    x : List, numpy array, or pandas series
        Numeric variable with only values 0 and 1
    na : bool
        Drop missing values before calculating (True or False)

    Returns
    -------
    float
        Calculated variance for a proportion based on a vector of 0 and 1 values

    Examples
    --------
    varprop([0, 1, 1, 1, 0, 0, 0])
    """

    p = ifelse(na, np.nanmean(x), np.mean(x))
    return p * (1 - p)


def seprop(x, na=True):
    """
    Calculate the standard error for a proportion

    Parameters
    ----------
    x : List, numpy array, or pandas series
        Numeric variable with only values 0 and 1
    na : bool
        Drop missing values before calculating (True or False)

    Returns
    -------
    float
        Calculated variance for a proportion based on a vector of 0 and 1 values

    Examples
    --------
    seprop([0, 1, 1, 1, 0, 0, 0])
    """
    x = np.array(x)
    if na:
        x = x[np.isnan(x) == False]
    return sqrt(varprop(x, na=False) / len(x))


def weighted_sd(df, wt):
    """
    Calculate the weighted standard deviation for a Pandas dataframe

    Parameters
    ----------
    df : Pandas dataframe
        All columns in the dataframe are expected to be numeric
    wt : List, pandas series, or numpy array
        Weights to use during calculation. The length of the vector should be the same as the number of rows in the df

    Returns
    -------
    Numpy array
        Array of weighted standard deviations for each column in df

    Examples
    --------
    df = pd.DataFrame({"x": [0, 1, 1, 1, 0, 0, 0]})
    wt = [1, 10, 1, 10, 1, 10, 1]
    weighted_sd(df, wt)
    """

    def wsd(x, wt):
        wt = wt / wt.sum()
        wm = np.average(x, axis=0, weights=wt)
        return sqrt((wt * (x - wm) ** 2).sum())

    wt = np.array(wt)
    return df.apply(lambda col: wsd(col, wt), axis=0).values


def weighted_mean(df, wt):
    """
    Calculate the weighted mean for a Pandas dataframe

    Parameters
    ----------
    df : Pandas dataframe
        All columns in the dataframe are expected to be numeric
    wt : List, pandas series, or numpy array
        Weights to use during calculation. The length of the vector should be the same as the number of rows in the df

    Returns
    -------
    Numpy array
        Array of weighted means for each column in df

    Examples
    --------
    df = pd.DataFrame({"x": [0, 1, 1, 1, 0, 0, 0]})
    wt = [1, 10, 1, 10, 1, 10, 1]
    weighted_mean(df, wt)
    """
    return np.average(df.values, weights=np.array(wt), axis=0)


def scale_df(df, wt=None, sf=2):
    """
    Scale the numeric variables in a Pandas dataframe

    Parameters
    ----------
    df : Pandas dataframe with numeric variables
    wt : Pandas series or None
        Weights to use during scaling. The length of the vector should be the same as the number of rows in the df
    sf : float
        Scale factor to use (default is 2)

    Returns
    -------
    Pandas dataframe with all numeric variables standardized

    Examples
    --------
    df = pd.DataFrame({"x": [0, 1, 1, 1, 0, 0, 0]})
    wt = [1, 10, 1, 10, 1, 10, 1]
    weighted_mean(df, wt)
    """
    df = df.copy()
    isNum = [pd.api.types.is_numeric_dtype(df[col]) for col in df.columns]
    isNum = list(compress(df.columns, isNum))
    dfs = df[isNum]
    if wt is None:
        df[isNum] = (dfs - dfs.mean().values) / (sf * dfs.std().values)
    else:
        df[isNum] = (dfs - weighted_mean(dfs, wt)) / (sf * weighted_sd(dfs, wt))
    return df



import math
import matplotlib.pyplot as plt
import pandas as pd
from pyrsm.utils import ifelse


def distr_plot(df):
    fig, axes = plt.subplots(
        math.ceil(df.shape[1] / 2), 2, figsize=(10, 1.5 * df.shape[1])
    )
    plt.subplots_adjust(wspace=0.25, hspace=0.3)
    row = 0
    for i, c in enumerate(df.columns):
        s = df[c]
        j = ifelse(i % 2 == 0, 0, 1)
        if pd.api.types.is_integer_dtype(s.dtype) and s.nunique() < 25:
            s.value_counts(sort=False).plot.bar(
                ax=axes[row, j], title=c, rot=0, color="slateblue"
            )
        elif pd.api.types.is_numeric_dtype(s.dtype):
            s.plot.hist(ax=axes[row, j], title=c, rot=0, color="slateblue")
        elif pd.api.types.is_categorical_dtype(s.dtype):
            s.value_counts(sort=False).plot.bar(
                ax=axes[row, j], title=c, rot=0, color="slateblue"
            )
        else:
            print(f"No plot for {c} (type {s.dtype})")

        if j == 1:
            row += 1



import pandas as pd
from scipy import stats


class cross_tabs:
    def __init__(self, df, var1, var2):
        self.df = df
        self.var1 = var1
        self.var2 = var2

        self.observed = pd.crosstab(
            df[var1], columns=df[var2], margins=True, margins_name="Total"
        )
        self.chisq = stats.chi2_contingency(
            self.observed.drop(columns="Total").drop("Total", axis=0), correction=False
        )
        expected = pd.DataFrame(self.chisq[3])
        expected["Total"] = expected.sum(axis=1)
        expected = expected.append(expected.sum(), ignore_index=True).set_index(
            self.observed.index
        )
        expected.columns = self.observed.columns
        self.expected = expected
        self.chi_sq = (
            ((self.observed - self.expected) ** 2 / self.expected)
            .drop(columns="Total")
            .drop("Total", axis=0)
        )
        self.perc_row = self.observed.div(self.observed["Total"], axis=0)
        self.perc_col = self.observed.div(self.observed.loc["Total", :], axis=1)
        self.perc = self.observed / self.observed.loc["Total", "Total"]

    def summary(self, output=["observed", "expected"], dec=2):
        prn = f"""
Cross-tabs
Data : titanic
Variables: survived, sex
Null hyp: there is no association between {self.var1} and {self.var2}
Alt. hyp: there is an association between {self.var1} and {self.var2}
"""
        if "observed" in output:
            prn = (
                prn
                + f"""
Observed:

{self.observed.applymap(lambda x: "{:,}".format(x))}
"""
            )
        if "observed" in output:
            prn = (
                prn
                + f"""
Expected: (row total x column total) / total

{self.expected.round(dec).applymap(lambda x: "{:,}".format(x))}
"""
            )
        if "chisq" in output:
            prn = (
                prn
                + f"""
Contribution to chi-squared: (o - e)^2 / e

{self.chi_sq.round(dec).applymap(lambda x: "{:,}".format(x))}
"""
            )
        if "perc_row" in output:
            prn = (
                prn
                + f"""
Row percentages:

{self.perc_row.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
            )
        if "perc_col" in output:
            prn = (
                prn
                + f"""
Column percentages:

{self.perc_col.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
            )

        if "perc_all" in output:
            prn = (
                prn
                + f"""
Percentages:

{self.perc.transform(lambda x: (100*x).round(dec).astype(str) + "%")}
"""
            )

        prn = (
            prn
            + f"""
Chi-squared: {round(self.chisq[0], dec)} df({round(self.chisq[2], dec)}), p.value {round(self.chisq[1], dec)}
"""
        )
        print(prn)

    # Another instance method
    def plot(self, output="perc_col"):
        pdf = getattr(self, output).drop(columns="Total").drop("Total", axis=0)
        fig = pdf.plot.bar()


def correlation(df, dec=3, prn=True):
    """
    Calculate correlations between the numeric variables in a Pandas dataframe

    Parameters
    ----------
    df : Pandas dataframe with numeric variables
    dec : int
        Number of decimal places to use in rounding
    prn : bool
        Print or return the correlation matrix

    Returns
    -------
    Pandas dataframe with all numeric variables standardized

    Examples
    --------
    df = pd.DataFrame({"x": [0, 1, 1, 1, 0], "y": [1, 0, 0, 0, np.NaN]})
    correlation(df)
    """
    df = df.copy()
    isNum = [pd.api.types.is_numeric_dtype(df[col]) for col in df.columns]
    isNum = list(compress(df.columns, isNum))
    df = df[isNum]

    ncol = df.shape[1]
    cr = np.zeros([ncol, ncol])
    cp = cr.copy()
    for i in range(ncol - 1):
        for j in range(i + 1, ncol):
            cdf = df.iloc[:, [i, j]]
            # pairwise deletion
            cdf = cdf[~np.any(np.isnan(cdf), axis=1)]
            c = stats.pearsonr(cdf.iloc[:, 0], cdf.iloc[:, 1])
            cr[j, i] = c[0]
            cp[j, i] = c[1]

    ind = np.triu_indices(ncol)

    # correlation matrix
    crs = cr.round(ncol).astype(str)
    crs[ind] = ""
    crs = pd.DataFrame(
        np.delete(np.delete(crs, 0, axis=0), crs.shape[1] - 1, axis=1),
        columns=df.columns[:-1],
        index=df.columns[1:],
    )

    # pvalues
    cps = cp.round(ncol).astype(str)
    cps[ind] = ""
    cps = pd.DataFrame(
        np.delete(np.delete(cps, 0, axis=0), cps.shape[1] - 1, axis=1),
        columns=df.columns[:-1],
        index=df.columns[1:],
    )

    if prn:
        print("Correlation matrix:")
        print(crs)
        print("\np.values:")
        print(cps)
    else:
        return cr, cp